# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import maya.OpenMayaMPx as omMPx
import maya.OpenMaya as om
import maya.mel as mel
import math
import random

kPluginNodeTypeName = "MASH_Spring"
mashSpringId = om.MTypeId(0x0011BE00)

class mashSpring(omMPx.MPxNode):

    outputArray = om.MObject()
    inputArray = om.MObject()
    aTime = om.MObject()
    aSpringAmount = om.MObject()
    aDamping = om.MObject()
    aStiffness = om.MObject()
    aStiffnessVar = om.MObject()
    aDampingVar = om.MObject()

    #AETemplate embedded in the .py file, get me.
    mel.eval('''
        global proc AEMASH_SpringTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Spring" "Spring";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "MASH Spring" -collapse 0;
        
        editorTemplate  -label "Spring Strength" -addControl "spring";

        editorTemplate -addSeparator;
        editorTemplate  -label "Damping" -addControl "damping";
        editorTemplate  -label "Stiffness" -addControl "stiffness";

        editorTemplate -addSeparator;
        editorTemplate  -label "Damping Variation" -addControl "dampingVariance";
        editorTemplate  -label "Stiffness Variation" -addControl "stiffnessVariance";
        
        editorTemplate -beginLayout "Inputs" -collapse 1;
        editorTemplate  -label "Time" -addControl "time";
        
        editorTemplate -suppress "inIterations";
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        ''')

    def __init__(self):
        omMPx.MPxNode.__init__(self)
        self._initialised = False
        self._previousTime = om.MTime()
        #make persistant position lists (and fill with nonsence to prevent Maya from boohooing)
        self._pyCurrentPosition = [0,0,0,0]
        self._pyPreviousPosition = [0,0,0,0]


    def compute(self, plug, data):
        if plug != mashSpring.outputArray:
            return om.kUnknownParameter    

        # Get slider values
        damping = data.inputValue(self.aDamping).asFloat()
        stiffness = data.inputValue(self.aStiffness).asFloat()
        dampingVariance = data.inputValue(self.aDampingVar).asFloat()
        stiffnessVariance = data.inputValue(self.aStiffnessVar).asFloat()
        springAmount = data.inputValue(self.aSpringAmount).asFloat()
        
        #Get the input array
        inputArray_dataHandle = data.inputValue(mashSpring.inputArray)
        #Get the output array
        outputArray_dataHandle = data.outputValue(mashSpring.outputArray)
        
        #Get the time
        currentTime = data.inputValue(self.aTime).asTime()
        
        #How many calculations to make
        numberOfCalcs = data.inputValue(mashSpring.inIterations).asInt()
                
        #make goal list and fill with nonsence, then empty
        goalArray = [0,0,0,0]
        del goalArray[:]
        
        #create output array
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        #Organise the input array
        inDArray = om.MVectorArray()
        inDArray.clear()
        inDArrData = data.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        #get the input array
        inDArray = inDArrFn.array()
        
        #organise the outputArray
        outDData = data.outputValue(self.outputArray)
        nData = om.MFnVectorArrayData()   
        
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
                
        #fill the goal array, but only if there's there is goal data...
        for h in range(numberOfCalcs):
            if (inDArray.length() > 0):
                testRead = om.MFloatVector(0.0,0.0,0.0)
                testRead = inDArray[h]

                goal = om.MFloatVector(testRead.x,testRead.y,testRead.z)
                goalArray.append(goal)
            
            # did you click the timeline and mess up the calculations? Reset!
            timeStep = currentTime.value() - self._previousTime.value()
            if timeStep > 1.0 or timeStep < 0.0:
                self._initialised = False
                self._previousTime = currentTime
                return
                
        #initialise the values if this is the first run, or if we need to because the timeline was clicked.
        if not self._initialised:
            #empty the lists
            del self._pyPreviousPosition[:]
            del self._pyCurrentPosition[:]
            del goalArray[:]
            
            for h in range(numberOfCalcs):
                #fill the lists!
                if (inDArray.length() > 0):
                    testRead = om.MFloatVector(0.0,0.0,0.0)
                    testRead = inDArray[h]
                    goal = om.MFloatVector(testRead.x,testRead.y,testRead.z)
                    goalArray.append(goal)
                    self._previousTime = currentTime
                    self._pyPreviousPosition.append(goalArray[h])
                    self._pyCurrentPosition.append(goalArray[h])
            
            self._initialised = True
        
        #This is where the shizzle happens...
        for h in range(numberOfCalcs):
            
            #use a seed to make the psudo random psudo.
            random.seed(h)
            
            #create unique random variances for each calculation
            dampingVarianceRand = dampingVariance * random.random()
            stiffnessVarianceRand = stiffnessVariance * random.random()

            #calculate the new position based on goals, previous positions, settings, alchemy and the position of the sun.
            dampingRand = damping * (1- dampingVarianceRand)
            stiffnessRand = stiffness * (1- stiffnessVarianceRand)
            velocity = (self._pyCurrentPosition[h] - self._pyPreviousPosition[h]) * (1.0 - dampingRand)
            newPosition = self._pyCurrentPosition[h] + velocity
            goalAttract = (goalArray[h] - newPosition) * stiffnessRand
            newPosition += goalAttract
            
            #set the lists for next time
            self._previousTime = om.MTime(currentTime)
            self._pyPreviousPosition[h] = self._pyCurrentPosition[h]
            self._pyCurrentPosition[h] = newPosition
            
            newPosition = goalArray[h] + ((newPosition - goalArray[h]) * springAmount)

            #add this result to the output array
            thisResult = om.MVector(newPosition.x,newPosition.y,newPosition.z)
            outResultArray.append(thisResult)

        #write the crap out of the output array  
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)
        #have a wash
        data.setClean(plug)


def nodeCreator():
    return omMPx.asMPxPtr(mashSpring())

def initialize():
    nAttr = om.MFnNumericAttribute()
    mAttr = om.MFnMatrixAttribute()
    uAttr = om.MFnUnitAttribute()
    
    #create a dummy array, for some reason, I don't quite know why, but it works.
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashSpring.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setReadable(True)
    tAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashSpring.addAttribute( mashSpring.inputArray)
    
    tAttr = om.MFnTypedAttribute()
    mashSpring.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(1)
    tAttr.setUsesArrayDataBuilder(True)
    mashSpring.addAttribute( mashSpring.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashSpring.inIterations = nAttr.create ( "inIterations", "inIter", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashSpring.addAttribute ( mashSpring.inIterations )

    mashSpring.aTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashSpring.addAttribute(mashSpring.aTime)
    mashSpring.attributeAffects(mashSpring.aTime, mashSpring.outputArray)

    mashSpring.aSpringAmount = nAttr.create("spring", "spring", om.MFnNumericData.kFloat, 1.0)
    nAttr.setKeyable(True)
    nAttr.setMin(0.0)
    nAttr.setMax(1.0)
    mashSpring.addAttribute(mashSpring.aSpringAmount)
    mashSpring.attributeAffects(mashSpring.aSpringAmount, mashSpring.outputArray)

    mashSpring.aStiffness = nAttr.create("stiffness", "stiff", om.MFnNumericData.kFloat, 0.2)
    nAttr.setKeyable(True)
    nAttr.setMin(0.0)
    nAttr.setMax(1.0)
    mashSpring.addAttribute(mashSpring.aStiffness)
    mashSpring.attributeAffects(mashSpring.aStiffness, mashSpring.outputArray)
    
    mashSpring.aStiffnessVar = nAttr.create("stiffnessVariance", "stiffVar", om.MFnNumericData.kFloat, 0.0)
    nAttr.setKeyable(True)
    nAttr.setMin(0.0)
    nAttr.setSoftMax(0.3)
    nAttr.setMax(1.0)
    mashSpring.addAttribute(mashSpring.aStiffnessVar)
    mashSpring.attributeAffects(mashSpring.aStiffnessVar, mashSpring.outputArray)
    
    mashSpring.aDampingVar = nAttr.create("dampingVariance", "dampVar", om.MFnNumericData.kFloat, 0.0)
    nAttr.setKeyable(True)
    nAttr.setMin(0.0)
    nAttr.setSoftMax(0.3)
    nAttr.setMax(1.0)
    mashSpring.addAttribute(mashSpring.aDampingVar)
    mashSpring.attributeAffects(mashSpring.aDampingVar, mashSpring.outputArray)

    mashSpring.aDamping = nAttr.create("damping", "damp", om.MFnNumericData.kFloat, 0.2)
    nAttr.setKeyable(True)
    nAttr.setMin(0.0)
    nAttr.setMax(1.0)
    mashSpring.addAttribute(mashSpring.aDamping)
    mashSpring.attributeAffects(mashSpring.aDamping, mashSpring.outputArray)

    mashSpring.attributeAffects( mashSpring.inputArray, mashSpring.outputArray)
    
def initializePlugin(obj):
    plugin = omMPx.MFnPlugin(obj, "Ian_Waters", "1.0", "Any")
    plugin.registerNode(kPluginNodeTypeName, mashSpringId, nodeCreator, initialize)


def uninitializePlugin(obj):
    plugin = omMPx.MFnPlugin(obj)
    plugin.deregisterNode(mashSpringId)

